/*

@author Sergio Alestra

*/

import ij.*;
import ij.gui.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import ij.text.*;

//plugin che si occupa di trovare i bordi

/*
IF d2  Zero AND d6  Zero THEN z5  Bianco
IF d6  Zero AND d8  Zero THEN z5  Bianco
IF d8  Zero AND d4  Zero THEN z5  Bianco
IF d4  Zero AND d2  Zero THEN z5  Bianco
                           ELSE z5  Nero
*/


public class Fuzzy_Find_Edge extends WindowAdapter implements PlugInFilter, ChangeListener, ActionListener{
  
  private int w;
  private int h;
  private ImagePlus im_ffe;
  private ImageProcessor ip_orig,ip_ffe;
  private byte[] im, imtmp;
  private ImageWindow w_ffe;
  
  //variabili per le regole
  private String[] rules;
  private TextWindow windowRules;
  
  //variabili per le funzioni di membership
  private float aZero = 0f;
  private float bZero = 64f;
  private float aWhite = 255f;
  private float bWhite = 190f;
  private float aBlack = 0f;
  private float bBlack = 190f;
  
  
  //slider per input mbf
  private JSlider s_aZero;
  private JSlider s_bZero;
  private JSlider s_aWhite;
  private JSlider s_bWhite;
  private JSlider s_aBlack;
  private JSlider s_bBlack;
  
  //etichette per mbf
  private JLabel l_aZero;
  private JLabel l_bZero;
  private JLabel l_aWhite;
  private JLabel l_bWhite;
  private JLabel l_aBlack;
  private JLabel l_bBlack;
  
  //bottonI applyZero e resetZero
  private JButton applyMbfZero,resetMbfZero;
  
  //bottone applyBW e resetBW
  private JButton applyMbfBW,resetMbfBW;
  
  //finestra per visualizzare la mbf Zero
  private JFrame windowMbfZero;
  private PlotPanel graphicMbfZero;
  
  //finestra per visualizzare le mbf Black and White
  private JFrame windowMbfBW;
  private PlotPanel graphicMbfBW;
  
  
  //metodo di inizializzazione richiesto da PlugInFilter
  public int setup(String arg, ImagePlus imp) {
    
    if (IJ.versionLessThan("1.37j"))
			return DONE;
		else
			return DOES_ALL+NO_CHANGES;
	}
  
  //metodo di esecuzione richiesto da PlugInFilter
  public void run(ImageProcessor ip) {
    
    w = ip.getWidth();
    h = ip.getHeight();
    ip_orig=ip;
    
    
    //creazione di una copia dell'immagine su cui lavorare
    im_ffe = NewImage.createByteImage("Fuzzy Find Edge",w,h,1,NewImage.FILL_BLACK);
    ip_ffe = (im_ffe.getProcessor()).convertToByte(true);
    ip_ffe.copyBits(ip,0,0,Blitter.COPY);
      
      
    //recuperiamo i valori dei pixels dell'immagine
    im = (byte[]) ip_ffe.getPixels();
    
    imtmp = new byte[im.length];
    for(int i=0; i<im.length; i++)
      imtmp[i]=im[i];
      
    //processiamo l'immagine
    this.process();
    
    //creiamo e visualizziamo le regole
    this.makeRules();
    this.showRules();
    
    //visualizziamo le mbf
    this.showMbfZero();
    this.showMbfBW();
    
    //visualizziamo l'immagine
    w_ffe = new ImageWindow(im_ffe);
    w_ffe.addWindowListener(this);
    im_ffe.updateAndDraw();
	}
  
  //-----------metodo makeRules-------------
  private void makeRules(){
    
    rules = new String[9];
    rules[0] = "IF d2 is zero AND d6 is zero THEN z5 is white";
    rules[1] = "IF d6 is zero AND d8 is zero THEN z5 is white";
    rules[2] = "IF d8 is zero AND d4 is zero THEN z5 is white";
    rules[3] = "IF d4 is zero AND d2 is zero THEN z5 is white";
    rules[4] = "                                             ELSE z5 is black";
    rules[5] = " ";
    rules[6] = "z1  z2  z3      d1 d2 d3";
    rules[7] = "z4  z5  z6  -> d4  0  d6      where d = z5 - z";
    rules[8] = "z7  z8  z9      d7 d8 d9";
  }
  
  //metodo per mostrare le regole
  private void showRules(){
    
    windowRules = new TextWindow("Rules - Fuzzy Find Edge",
                                  "", 500, 350);
    
    for(int i=0; i<rules.length; i++)
      windowRules.append(rules[i]);
    
    windowRules.setVisible(true);
  }
  
  //---------metodo process--------------
  private void process(){
    
    int pos, offset, int_im_pos;
    int sx, dx, up, down;
    int d2, d4, d6, d8;
    
    float d2_fuzzy, d4_fuzzy, d6_fuzzy, d8_fuzzy;
    float ant1, ant2, ant3, ant4;
    
    //scandisco tutti i pixels
    for(int i = 0; i < h; i++){
      
      offset = i*w;
      for(int j = 0; j < w; j++){
        
        pos = offset+j;
        
        //cast ad int del valore di input del pixel
        int_im_pos = 0xff & im[pos];
        
        //calcolo dei pixels vicini e cast a int
        sx =   0xff & ((pos-1)<0 ? 0 : im[pos-1]);                //pixel di sinistra
        dx =   0xff & ((pos+1)>=im.length ? 0 : im[pos+1]);         //pixel di destra
        up =   0xff & (((i-1)*w+j)<0 ? 0 : im[(i-1)*w+j]);        //pixel di sopra
        down = 0xff & (((i+1)*w+j)>=im.length ? 0 : im[(i+1)*w+j]); //pixel di sotto
        
        //calcolo delle differenze con i pixels vicini
        d2 = up-int_im_pos;
        d4 = sx-int_im_pos;
        d6 = dx-int_im_pos;
        d8 = down-int_im_pos;
        
        //FUZZYFICAZIONE delle differenze
        d2_fuzzy = this.mbfZero(d2);
        d4_fuzzy = this.mbfZero(d4);
        d6_fuzzy = this.mbfZero(d6);
        d8_fuzzy = this.mbfZero(d8);
        
        //Applicazione degli operatori fuzzy
        ant1=Math.min(d2_fuzzy, d6_fuzzy);  //IF d2  Zero AND d6  Zero
        ant2=Math.min(d6_fuzzy, d8_fuzzy);  //IF d6  Zero AND d8  Zero
        ant3=Math.min(d8_fuzzy, d4_fuzzy);  //IF d8  Zero AND d4  Zero
        ant4=Math.min(d4_fuzzy, d2_fuzzy);  //IF d4  Zero AND d2  Zero
        
        //IMPLICAZIONE, AGGREGAZIONE, calcolo ascissa del centroide
        
        int_im_pos = (int) xAxisCentroid(255, ant1, ant2, ant3, ant4);
        
        imtmp[pos] = (byte) int_im_pos;
      }
    }
    
    for(int i=0; i<im.length; i++)
      im[i]=imtmp[i];
  }
   
  
  //funzione di Membership Zero di tipo Bell-Shape
  private float mbfZero(int px){
    
    if(px <= aZero)
      return S_Shape(px, aZero-bZero, aZero-bZero/2, aZero);
    else
      return 1f-S_Shape(px, aZero, aZero+bZero/2, aZero+bZero);
  }
  
  //funzione di Membership di tipo S-Shape utilizzata da mbfZero
  private float S_Shape(int px, float a, float b, float c){
    
    if(px < a)
      return 0f;
    else if( (px >= a) && (px<=b) )
      return 2f*(float)Math.pow((px-a)/(c-a),2);
    else if( (px > b) && (px <= c) )
      return 1-2f*(float)Math.pow((px-c)/(c-a),2);
    else
      return 1f;
  }
  
  //funzione di Membership Nero di tipo Sigma
  private float mbfBlack(int px){
    
    if( (px >= (aBlack)) && (px <= (aBlack+bBlack)))
      return 1-(px-aBlack)/bBlack;
    else if( px < aBlack )
      return 1;
    else
      return 0;
  }
  
  //funzione di Membership Bianco di tipo Sigma
  private float mbfWhite(int px){
    
    if( (px >= (aWhite-bWhite)) && (px < aWhite))
      return 1-(aWhite-px)/bWhite;
    else if( px >= aWhite )
      return 1;
    else
      return 0;
  }
  
  //calcolo dell'ascissa del centroide attraverso il calcolo 
  //dell'integrale approssimato con la regola del Trapezio
  //gaps (numero di intervalli)
  
  private float xAxisCentroid(int gaps, float ant1, float ant2, float ant3, float ant4){
    
    float[] x = new float[gaps+1];
    
    for(int i=0; i<gaps+1; i++)
      x[i]=i*255/gaps;
    
    float numerator = 0;
    float denominator = 0;
    
    float f_a = 0, f_b = 0;
    float f_a1 = 0, f_a2 = 0, f_a3 = 0, f_a4 = 0;
    float f_b1 = 0, f_b2 = 0, f_b3 = 0, f_b4 = 0;
    float f_a_else = 0, f_b_else = 0;
    
    for(int i =0; i<gaps-1; i++){
      
      //estremi del sottointervallo di integrazione
      f_a = mbfWhite((int)x[i]);
      f_b = mbfWhite((int)x[i+1]);
      
      //IMPLICAZIONE (operatore di min)
      
      f_a1 = Math.min(ant1,f_a);
      f_a2 = Math.min(ant2,f_a);
      f_a3 = Math.min(ant3,f_a);
      f_a4 = Math.min(ant4,f_a);
      
      f_b1 = Math.min(ant1,f_b);
      f_b2 = Math.min(ant2,f_b);
      f_b3 = Math.min(ant3,f_b);
      f_b4 = Math.min(ant4,f_b);
      
      //ELSE
      
      f_a_else = 0.5f*Math.min(1-f_a1, Math.min(1-f_a2, Math.min(1-f_a3, 1-f_a4)));
      f_b_else = 0.5f*Math.min(1-f_b1, Math.min(1-f_b2, Math.min(1-f_b3, 1-f_b4)));
      
      
      //AGGREGAZIONE (operatore di max)
      
      f_a = Math.max(f_a_else, Math.max(f_a1, Math.max(f_a2, Math.max(f_a3, f_a4))));
      f_b = Math.max(f_b_else, Math.max(f_b1, Math.max(f_b2, Math.max(f_b3, f_b4))));
      
      //numeratore e denominatore per calcolo dell'ascissa del centroide
      
      numerator += (x[i+1] - x[i])/2 *(x[i]*f_a + x[i+1]*f_b);
      denominator += (x[i+1] - x[i])/2 *(f_a + f_b);
      
    }
    
    //DEFUZZTFICAZIONE (calcolo ascissa del centroide)
    
    if(numerator/denominator > 127)
      return 255;
    else
      return (numerator/denominator);
    
  }
  
  //metodo per plottare l' mbf Zero
  private ImagePlus plotMbfZero(){
    
    
    float[] xZ = new float[510];
    float[] yZ = new float[510];
    
    
    for(int i =0; i<510; i++){
      xZ[i]=i-255;
      yZ[i]=mbfZero(i-255);
    }
    
    //mbf Zero
    Plot p = new Plot("Membership function Zero - Fuzzy Find Edge",
                                    "","",xZ,yZ);
    p.setLimits(-255.0,255.0,0.0,1.2);
    p.setLineWidth(2);
    
    //pw.addLabel(0.1,0.1,"mbf Zero");
    
    return p.getImagePlus();
  }
  
  //metodo per plottare le mbf Black and White
  private ImagePlus plotMbfBW(){
    
    
    float[] xBW = new float[256];
    float[] yB = new float[256];
    float[] yW = new float[256];
    
    for(int i =0; i<256; i++){
      xBW[i]=i;
      yB[i]=mbfBlack(i);
      yW[i]=mbfWhite(i);
    }
    
    //mbf Black e White
    Plot p = new Plot("Membership function Black and White - Fuzzy Find Edge",
                                    "","",xBW,yB);
    p.setLimits(0,255.0,0.0,1.2);
    p.setLineWidth(2);
    p.addPoints(xBW,yW,PlotWindow.LINE);
    
    //pw2.addLabel(0.1,0.1,"mbf Black");
    //pw2.addLabel(0.8,0.1,"mbf White");
    
    return p.getImagePlus();
  }
  
  //--------------metodi per la finestra delle MbfZero
  
  private void showMbfZero(){
    
    //creo i bottoni
    applyMbfZero = new JButton("Apply");
    applyMbfZero.addActionListener(this);
    resetMbfZero = new JButton("Reset");
    resetMbfZero.addActionListener(this);
    
    //creo i pannelli
    JPanel panelZero = new JPanel(new GridLayout(2,2));
    JPanel panelApply_Reset = new JPanel();
    graphicMbfZero = new PlotPanel(this.plotMbfZero().getImage());
    
    //aggiungo un bordo ai pannelli
    panelZero.setBorder(BorderFactory.createTitledBorder("Membership function Zero"));
    
    //creo le etichette
    l_aZero = new JLabel();
    l_bZero = new JLabel();
    
    this.setLabelMbfZero();
    
    //aggiungo le etichette dei campi
    panelZero.add(l_aZero);
    panelZero.add(l_bZero);
    
    //creo gli slider
    s_aZero = new JSlider(JSlider.HORIZONTAL);
    s_bZero = new JSlider(JSlider.HORIZONTAL);
    
    this.setSliderMbfZero();
    
    s_aZero.addChangeListener(this);
    s_bZero.addChangeListener(this);
    
    //aggiungo gli slider
    panelZero.add(s_aZero);
    panelZero.add(s_bZero);
    panelApply_Reset.add(applyMbfZero);
    panelApply_Reset.add(resetMbfZero);
    
    //creo la finestra per le mbf
    windowMbfZero = new JFrame("Membership function Zero - Fuzzy Find Edge");
    windowMbfZero.setSize(700,450);
    windowMbfZero.setLocation(300,200);
    windowMbfZero.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    windowMbfZero.setLayout(new FlowLayout());
    Container contentPane=windowMbfZero.getContentPane();
    
    contentPane.add(graphicMbfZero);
    contentPane.add(panelZero);
    contentPane.add(panelApply_Reset);
    
    windowMbfZero.setVisible(true);
  }
  
  private void setLabelMbfZero(){
    
    l_aZero.setText("a = "+(int)aZero);
    l_bZero.setText("b = "+(int)bZero);
  }
  
  private void setSliderMbfZero(){
    
    s_aZero.setMinimum(-255+(int)bZero);
    s_aZero.setMaximum(255-(int)bZero);
    s_aZero.setValue((int)aZero);
    s_bZero.setMinimum(0);
    s_bZero.setMaximum(255-Math.abs((int)aZero));
    s_bZero.setValue((int)bZero);
  }
  
  private void resetMbfZero(){
    
    aZero = 0f;
    bZero = 64f;
    s_aZero.setMinimum(-255+(int)bZero);
    s_aZero.setMaximum(255-(int)bZero);
    s_aZero.setValue(0);
    s_bZero.setMinimum(0);
    s_bZero.setMaximum(255-Math.abs((int)aZero));
    s_bZero.setValue(64);
  }
  
  //----------------metodi per la finestra delle mbf Black and White---------------
  
  private void showMbfBW(){
    
    //creo i bottoni
    applyMbfBW = new JButton("Apply");
    applyMbfBW.addActionListener(this);
    resetMbfBW = new JButton("Reset");
    resetMbfBW.addActionListener(this);
    
    //creo i pannelli
    JPanel panelWhite = new JPanel(new GridLayout(2,2));
    JPanel panelBlack = new JPanel(new GridLayout(2,2));
    JPanel panelApply_Reset = new JPanel();
    graphicMbfBW = new PlotPanel(this.plotMbfBW().getImage());
    
    //aggiungo un bordo ai pannelli
    panelWhite.setBorder(BorderFactory.createTitledBorder("Membership function White"));
    panelBlack.setBorder(BorderFactory.createTitledBorder("Membership function Black"));
    
    //creo le etichette
    l_aWhite = new JLabel();
    l_bWhite = new JLabel();
    l_aBlack = new JLabel();
    l_bBlack = new JLabel();
    
    this.setLabelMbfBW();
    
    //aggiungo le etichette dei campi
    panelWhite.add(l_aWhite);
    panelWhite.add(l_bWhite);
    panelBlack.add(l_aBlack);
    panelBlack.add(l_bBlack);
    
    //creo gli slider
    s_aWhite = new JSlider(JSlider.HORIZONTAL);
    s_bWhite = new JSlider(JSlider.HORIZONTAL);
    s_aBlack = new JSlider(JSlider.HORIZONTAL);
    s_bBlack = new JSlider(JSlider.HORIZONTAL);
    
    this.setSliderMbfBW();
    
    s_aWhite.addChangeListener(this);
    s_bWhite.addChangeListener(this);
    s_aBlack.addChangeListener(this);
    s_bBlack.addChangeListener(this);
    
    //aggiungo gli slider
    panelWhite.add(s_aWhite);
    panelWhite.add(s_bWhite);
    panelBlack.add(s_aBlack);
    panelBlack.add(s_bBlack);
    panelApply_Reset.add(applyMbfBW);
    panelApply_Reset.add(resetMbfBW);
    
    //creo la finestra per le mbf
    windowMbfBW = new JFrame("Membership function Black and White - Fuzzy Find Edge");
    windowMbfBW.setSize(700,500);
    windowMbfBW.setLocation(300,200);
    windowMbfBW.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    windowMbfBW.setLayout(new FlowLayout());
    Container contentPane=windowMbfBW.getContentPane();
    
    contentPane.add(graphicMbfBW);
    contentPane.add(panelWhite);
    contentPane.add(panelBlack);
    contentPane.add(panelApply_Reset);
    
    windowMbfBW.setVisible(true);
  }
  
  private void setLabelMbfBW(){
    
    l_aWhite.setText("a = "+(int)aWhite);
    l_bWhite.setText("b = "+(int)bWhite);
    l_aBlack.setText("a = "+(int)aBlack);
    l_bBlack.setText("b = "+(int)bBlack);
  }
  
  private void setSliderMbfBW(){
    
    s_aBlack.setMinimum(0);
    s_aBlack.setMaximum(255-(int)bBlack);
    s_aBlack.setValue((int)aBlack);
    s_bBlack.setMinimum(0);
    s_bBlack.setMaximum(255-(int)aBlack);
    s_bBlack.setValue((int)bBlack);
    
    s_aWhite.setMinimum(0+(int)bWhite);
    s_aWhite.setMaximum(255);
    s_aWhite.setValue((int)aWhite);
    s_bWhite.setMinimum(0);
    s_bWhite.setMaximum((int)aWhite);
    s_bWhite.setValue((int)bWhite);
  }
  
  private void resetMbfBW(){
    
    aWhite = 255f;
    bWhite = 190f;
    aBlack = 0f;
    bBlack = 190f;
    
    s_aBlack.setMinimum(0);
    s_aBlack.setMaximum(255-(int)bBlack);
    s_aBlack.setValue(0);
    s_bBlack.setMinimum(0);
    s_bBlack.setMaximum(255-(int)aBlack);
    s_bBlack.setValue(190);
    
    s_aWhite.setMinimum(0+(int)bWhite);
    s_aWhite.setMaximum(255);
    s_aWhite.setValue(255);
    s_bWhite.setMinimum(0);
    s_bWhite.setMaximum((int)aWhite);
    s_bWhite.setValue(190);
  }
  
  //metodi per il controllo degli eventi
  public void stateChanged(ChangeEvent e){
    
    
    Object source=e.getSource();
    
    if(source==s_aZero){
      
      aZero = s_aZero.getValue();
      
      setSliderMbfZero();
      setLabelMbfZero();
      graphicMbfZero.updateImage(plotMbfZero().getImage());
    }
    
    if(source==s_bZero){
      
      bZero = s_bZero.getValue();
      
      setSliderMbfZero();
      setLabelMbfZero();
      graphicMbfZero.updateImage(plotMbfZero().getImage());
    }
    
    if(source==s_aBlack){
      
      aBlack = s_aBlack.getValue();
      
      setSliderMbfBW();
      setLabelMbfBW();
      graphicMbfBW.updateImage(plotMbfBW().getImage());
    }
    
    if(source==s_bBlack){
      
      bBlack = s_bBlack.getValue();
      
      setSliderMbfBW();
      setLabelMbfBW();
      graphicMbfBW.updateImage(plotMbfBW().getImage());
    }
    
    if(source==s_aWhite){
      
      aWhite = s_aWhite.getValue();
      
      setSliderMbfBW();
      setLabelMbfBW();
      graphicMbfBW.updateImage(plotMbfBW().getImage());
    }
    
    if(source==s_bWhite){
      
      bWhite = s_bWhite.getValue();
      
      setSliderMbfBW();
      setLabelMbfBW();
      graphicMbfBW.updateImage(plotMbfBW().getImage());
    }
    
  }
  
  public void actionPerformed(ActionEvent e){
    
    Object source=e.getSource();
    
    if(source==applyMbfZero){
      ip_ffe.copyBits(ip_orig,0,0,Blitter.COPY);
      im = (byte[]) ip_ffe.getPixels();
      for(int i=0; i<im.length; i++)
        imtmp[i]=im[i];
      
      this.process();
      im_ffe.updateAndDraw();
    }
    
    if(source==resetMbfZero){
      this.resetMbfZero();
      setLabelMbfZero();
      graphicMbfZero.updateImage(plotMbfZero().getImage());
    }
    
    if(source==applyMbfBW){
      ip_ffe.copyBits(ip_orig,0,0,Blitter.COPY);
      im = (byte[]) ip_ffe.getPixels();
      for(int i=0; i<im.length; i++)
        imtmp[i]=im[i];
      
      this.process();
      im_ffe.updateAndDraw();
    }
    
    if(source==resetMbfBW){
      this.resetMbfBW();
      setLabelMbfBW();
      graphicMbfBW.updateImage(plotMbfBW().getImage());
    }
    
  }
  
  public void windowClosing(WindowEvent e){
    
    windowMbfZero.setVisible(false);
    windowMbfBW.setVisible(false);
    windowRules.setVisible(false);
  }
}